#pragma once

#include <glib-object.h>

G_BEGIN_DECLS

typedef struct _OTPClientApplication OTPClientApplication;

typedef struct _OTPClientWindow OTPClientWindow;

typedef struct _OTPEntry OTPEntry;

G_END_DECLS